// Viene inclusa la libreria LIQUID_CRYSTAL con 
// tutte le sue funzioni
//
#include <LiquidCrystal.h>

// TIMER LCD
//
// Esercitazione con display LCD 
// Timer 0.5...99.5 secondi
//
// ITS AUTOMA-20
//
// Lezione 5: 28/06/2021
//

// Definizione ingressi
#define PORTA_PULS_PIU	  11
#define PORTA_PULS_MENO	  10
#define PORTA_PULS_START   9
#define PORTA_PULS_STOP   12
// Definizione uscite (per LCD)
#define LCD_RS	8
#define LCD_EN  7
#define LCD_DB4 6
#define LCD_DB5 5
#define LCD_DB6 4
#define LCD_DB7 3
// Altre uscite
#define OUT_MOTOR  13
// Definizione stato attivo ingressi
#define IN_ATTIVO  		1
#define IN_NON_ATTIVO 	0
// Definizione stato attivo uscite
#define OUT_ATTIVA   	1
#define OUT_NON_ATTIVA 	0
// Definisco anche il numero di ordine a cui corrispondono
// i vari pulsanti, ricordarsi che sono tre pulsanti e la
// numerazione parte sempre da zero nel linguaggio C
#define PULS_PIU    0
#define PULS_MENO   1
#define PULS_START  2
#define PULS_STOP   3
#define NUM_PULSANTI 4

int TempoMancante;
int TempoImpostato;
bool OutMotor;
bool PauseMotor;
int Secondi, oldSecondi;
int Decimi, Centesimi;
unsigned long long TicksMS;
// Definizione di una struttura necessaria a contenere
// tutti i dati relativi ai pulsanti, si tratta a tutti
// gli effetti della definizione di un contenitore sui cui
// si appoggiano, isolandondoli dal resto delle variabili,
// solo le variabili relative ai pulsanti
typedef struct 
{
    int Porta;
    bool newLettura;
    bool oldLettura;
    bool StatoAttuale;
    bool StatoPrecedente;
    bool AutoRepeat;
    unsigned long long TimeUltimaVariazione;  
    unsigned long long TimePressione;
} _PULS_STRUCTURE;

// A questo punto definisco un array di 'n' valori (poichè
// sono 'n' pulsanti) con la struttura appena definita

_PULS_STRUCTURE Pulsante[NUM_PULSANTI];

// Definisce (inizializza i piedini utilizzati dal display LCD
// facendo riferimento alla libreria indicata dal file di Header)
// Sintassi: LiquidCrystal <variabile>(rs, enable, d4, d5, d6, d7)
LiquidCrystal lcd(LCD_RS, LCD_EN, LCD_DB4, LCD_DB5, LCD_DB6, LCD_DB7);
// Notare che si poteva scrivere, anche se in maniera più criptica
// LiquidCrystal lcd(3,4,5,6,7,8);

void setup()
{
  int i;
  Serial.begin(115200);
  // Definizione pin di ingresso
  pinMode(PORTA_PULS_PIU, INPUT);
  pinMode(PORTA_PULS_MENO, INPUT);
  pinMode(PORTA_PULS_START, INPUT);
  pinMode(PORTA_PULS_STOP, INPUT);
  // Definizione pin di uscita
  pinMode(OUT_MOTOR, OUTPUT); 
  // NOTARE CHE non importa definire come uscite
  // i piedini utilizzati per il display, a ciò provvedono
  // opportune funzioni nella libreria LIQUID CRYSTAL
  //
  // Inizializzazione variabili del programma
  OutMotor = false;
  PauseMotor = false;
  // Per default parte con 5.0 secondi ovvero 50 decimi
  TempoImpostato = 50;
  TempoMancante = 0;
  TicksMS = millis() + 100;
  Secondi = oldSecondi = 0;
  // Azzera tutte le variabili relative a tutti i pulsanti
  for( i=0; i<NUM_PULSANTI; i++ )
  {
    Pulsante[i].newLettura = false;
    Pulsante[i].oldLettura = false;
    Pulsante[i].StatoAttuale = false;
    Pulsante[i].StatoPrecedente = false;
    Pulsante[i].AutoRepeat = false;
    Pulsante[i].TimeUltimaVariazione = millis();
    Pulsante[i].TimePressione = millis();
  }
  // Inserisce la porta (o piedino) a cui è collegato
  // ogni pulsante
  Pulsante[PULS_PIU].Porta = PORTA_PULS_PIU;
  Pulsante[PULS_MENO].Porta = PORTA_PULS_MENO;
  Pulsante[PULS_START].Porta = PORTA_PULS_START;
  Pulsante[PULS_STOP].Porta = PORTA_PULS_STOP;
  // Attiva la funzionalità AutoRepeat solo sui due
  // pulsanti PIU e MENO
  Pulsante[PULS_PIU].AutoRepeat = true;
  Pulsante[PULS_MENO].AutoRepeat = true;  
  // Azzera tutte le uscite
  digitalWrite( OUT_MOTOR, LOW );
  // Preparo il display LCD
  // Inizializzazione della libreria LCD
  lcd.begin(16,2);  
  // Cancellazione completa del contenuto (in caratteri)
  lcd.clear();
  // Menù di benvenuto
  lcd.setCursor( 0, 0 ); // Si posiziona a COL 0, ROW 0
  // --------0123456789012345-------------------------
  lcd.print("Timer 99 secondi");
  lcd.setCursor( 0, 1 ); // Si posiziona a COL 0, ROW 1
  // --------0123456789012345-------------------------
  lcd.print("by ITS AUTOMA-20");  
  // Attende 2 secondi
  delay(2000);
  // Cancella tutto il display per poter scrivere il
  // menù principale
  lcd.clear();
  // Menù principale
  lcd.setCursor( 0, 0 ); // Si posiziona a COL 0, ROW 0
  // --------0123456789012345-------------------------
  lcd.print("Set Time: __._ s");
  lcd.setCursor( 0, 1 ); // Si posiziona a COL 0, ROW 1
  // --------0123456789012345-------------------------
  lcd.print("MOTOR STOPPED   ");    
}

void loop()
{
  int i;
  // Controllo con timer di sistema con precisione di
  // circa 1 millisecondo
  if( millis() > TicksMS )
  {
    // Sono passati 100 millisecondi ovvero 1/10 secondo
    // incrementa il contatore dei DECIMI
    ++Decimi;
    // Ricarica timer per prossimo incremento dei secondi
    TicksMS = millis() + 100;
    // Refresh del display ogni 100 msec
    RefreshDisplay();
    // Se la temporizzazione era attiva allora scala il
    // tempo rimanente (fino ad arrivare a zero, NON deve
    // scendere sotto lo zero), inoltre quando il tempo
    // residuo arriva a zero spenge anche il relè
    if( OutMotor == true )
    {
      if( TempoMancante > 0 ) 
      {
        if( ! PauseMotor ) --TempoMancante;
      }
      else
      {
        OutMotor = false;
      }
    }
    else
    {
      // Altrimenti non stava temporizzando e allora non fa nulla
    }  
  }  
  
  // Lettura dei pulsanti di ingresso
  // Per evitare di ingigantire troppo il MAIN LOOP
  // il controllo viene fatto in una funzione appositamente
  // progettata
  for( i=0; i<NUM_PULSANTI; i++ ) LeggePulsante(i);
  
  // Verifica se qualche situazione è cambiata e prende decisioni
  // in merito (se il timer era in moto non considera la pressione
  // dei pulsanti e quindi non modificherà nulla)
  if( OutMotor == false )
  {
    // Si poteva scrivere anche in un modo più compatto (vediamo la stessa cosa
    // per il controllo sul Pulsante PIU (incremento tempo)
    if(    (Pulsante[PULS_PIU].StatoAttuale == true)
        && (Pulsante[PULS_PIU].StatoAttuale !=  Pulsante[PULS_PIU].StatoPrecedente )
       )
    {
      // 
      if( TempoImpostato <= 990 ) TempoImpostato += 5;
    }
    // Stesso controllo per Pulsante MENO (decremento tempo)
    // ...ancora una volta si può incorporare anche il terzo
    // controllo nello stesso IF
    if(    (Pulsante[PULS_MENO].StatoAttuale == true)
        && (Pulsante[PULS_MENO].StatoAttuale !=  Pulsante[PULS_MENO].StatoPrecedente )
        && (TempoImpostato > 5)
       )
    {
      // 
      TempoImpostato -= 5;
    }
  }
  
  // Gestione del pulsante di STOP
  // Semplicemente viene azzerato immediatamente il tempo mancante
  // provocando quindi lo spengimento del motore
  if(     (Pulsante[PULS_STOP].StatoAttuale != Pulsante[PULS_STOP].StatoPrecedente)
       && (Pulsante[PULS_STOP].StatoAttuale == true)
       && (OutMotor == true)
     )
     TempoMancante = 0; 
  
  // E' stato premuto il pulsante START (adesso è TRUE e prima
  // era FALSE), acquisisce il tempo impostato attuale come tempo
  // rimanente e fa partire il timer
  if( Pulsante[PULS_START].StatoAttuale != Pulsante[PULS_START].StatoPrecedente )
  {
    // Considera solo l'evento della pressione del pulsante
    // e NON il suo rilascio
    if( Pulsante[PULS_START].StatoAttuale == true )
    {
      if( OutMotor == false )
      {
        // Non era partita la temporizzazione
        // allora la fa partire con tempo pieno
        TempoMancante = TempoImpostato;
        PauseMotor = false;
        OutMotor = true;
      }
      else
      {
        // Una volta partita la temporizzazione la pressione del 
        // pulsante di START provoca alternativamente la PAUSA
        // e lo START del motore (senza ripartire dall'inizio)
        if( PauseMotor ) PauseMotor = false; else PauseMotor = true;
      }
    }
  }  
  // Dopo le verifiche porta il valore StatoPrecedente al valore StatoAttuale
  // in modo da NON riprendere in considerazione gli stessi eventi
  // (lo fa per TUTTI i PULSANTI)
  for( i=0; i<NUM_PULSANTI; i++ ) Pulsante[i].StatoPrecedente = Pulsante[i].StatoAttuale;
  
  // *********************************************************************************
  // Visualizzazione su display LCD
  // *********************************************************************************
  // Non essendo necessaria il rinfresco del display ad ogni "giro"
  // della funzione LOOP si consiglia di porre questa funzione all'interno
  // del controllo IF che viene eseguito ogni 100 msec che è un tempo
  // più che sufficiente per gli "umani" che non possono apprezzare variazioni
  // più veloci della visualizzazione
  // Per cui commentiamo la funzione in questo punto e la copiamo più in alto
  // RefreshDisplay();
  
  // Imposta correttamente l'uscita collegata al relè
  if( OutMotor & ! PauseMotor )
  	digitalWrite( OUT_MOTOR, OUT_ATTIVA );
  else
    digitalWrite( OUT_MOTOR, OUT_NON_ATTIVA );
}

void RefreshDisplay(void)
{
  int Tempo, Secondi, Decimi;
  //
  // In base allo stato del motore vengono modificate alcune
  // scritte sul menù principale del display
  //
  if( OutMotor == false )
  {
    // TEMPORIZZAZIONE FERMA ***************************
    Tempo = TempoImpostato;
    // Menù principale
    lcd.setCursor( 0, 0 ); // Si posiziona a COL 0, ROW 0
    // --------0123456789012345-------------------------
    lcd.print("Set ");  
    lcd.setCursor( 6, 1 ); // Si posiziona a COL 0, ROW 1
    // --------0123456789012345-------------------------  
    lcd.print("STOPPED   "); 
  }
  else
  {
    // TEMPORIZZAZIONE IN CORSO ************************
    Tempo = TempoMancante;
    // Menù principale
    lcd.setCursor( 0, 0 ); // Si posiziona a COL 0, ROW 0
    // --------0123456789012345-------------------------
    lcd.print("Rem.");
    lcd.setCursor( 6, 1 ); // Si posiziona a COL 0, ROW 1
    // --------0123456789012345-------------------------  
    // ***** Eventuale PAUSA del motore
    if( PauseMotor )
      lcd.print("PAUSED    ");  
    else
      lcd.print("RUNNING...");     
  }
  // ************************************
  // Suddivide il tempo espresso in decimi di secondo
  // nelle due parti SECONDI e DECIMI per poterlo
  // visualizzare come intero con decimali
  //
  // Parte intera del tempo (secondi)
  Secondi = Tempo / 10;
  // Resto della divisione
  // Decimi = Tempo - (Secondi*10);
  Decimi = Tempo % 10;
  // Scrivo il tempo in secondi (interi)
  lcd.setCursor(10,0); // Si posiziona a COL 10, ROW 0 
  // Se deve scrivere solo unità allora cancella il
  // dato presente nelle decine (scrive uno SPAZIO)
  if( Secondi < 10 ) 
  {
  	lcd.print(" ");  
  }
  lcd.print(Secondi);
  // Scrivo il tempo in decimi 
  lcd.setCursor(13,0); // Si posiziona a COL 13, ROW 0 
  lcd.print(Decimi);
}

void LeggePulsante( int pu )
{
    // Evita di fare qualsiasi cosa se viene passato
    // un numero inesistente di pulsante
    if( pu < NUM_PULSANTI )
    {
      // Legge lo stato attuale dell'ingresso
      Pulsante[pu].newLettura = digitalRead(Pulsante[pu].Porta);
      // Confronto con lo stato precedente (se è cambiato o no)
      if( Pulsante[pu].newLettura != Pulsante[pu].oldLettura )
      {
        // Valore cambiato, carica il tempo di conteggio
        // di timeout (acquisizione del valore)
        #define DEBOUNCE_TIME      20
        Pulsante[pu].TimeUltimaVariazione = millis() + DEBOUNCE_TIME;
        // memorizza il valore come OLD per prossimo confronto
        Pulsante[pu].oldLettura = Pulsante[pu].newLettura ;
        // Se il pulsante è premuto inizializza il tempo di AutoRepeat
        // Questo è il tempo di ritardo per l'attivazione, teoricamente
        // dovrebbe essere impostato a 1 secondo ma, simulandolo su 
        // TinkerCad che è molto lento, conviene ridurlo...
        #define AUTO_REPEAT_TIME  100
        Pulsante[pu].TimePressione = millis() + AUTO_REPEAT_TIME;
      }
      else
      {
        // Valore NON cambiato rispetto a prima, allora se è
        // trascorso il tempo di antirimbalzo acquisisce il 
        // valore come StatoAttuale del pulsante
        if( millis() > Pulsante[pu].TimeUltimaVariazione )
        {
          Pulsante[pu].StatoAttuale = Pulsante[pu].newLettura;
        } 
        // Controllo se il pulsante è premuto per gestire il
        // sistema di AUTOREPEAT
        if( Pulsante[pu].StatoAttuale == IN_ATTIVO )
        {
          if( millis() > Pulsante[pu].TimePressione )
          {
            if( Pulsante[pu].AutoRepeat == true )
            {        
              // Questo rappresenta il tempo di AutoRepeat, dovrebbe
              // essere di 200 msec ma, in ambiente TinkerCad, è bene
              // ridurlo notevolmente...
              #define REPEAT_TIME  20
        	  Pulsante[pu].TimePressione = millis() + REPEAT_TIME;  
              // Simula il cambiamento di stato del pulsante, modificando
              // artificiosamente lo stato precedente
              Pulsante[pu].StatoPrecedente = IN_NON_ATTIVO;
            }
          }         
        }
      }  
    }
}




